import logging
from ..exceptions import NoTXFound, NoTXToProcess, NumConfirmationsNotMet

from requests.exceptions import HTTPError
from urllib3.exceptions import NewConnectionError

from ..configuration import ServerConfig
from ..factories.transaction_factory import TransactionFactory
from ..transactions.monero_transaction import MoneroTransaction


class MoneroFactory(TransactionFactory):
    payment_provider = "Monero"
    server_config: ServerConfig = None

    # you would specify no_data is True when you only care
    # about the fact that a transaction has been received
    # and you don't care about the details of that transaction
    def get_transaction(
        self, tx_id: str, get_tx_data: bool, get_raw_data: bool = False
    ) -> MoneroTransaction:
        tx = MoneroTransaction(
            server_config=self.server_config,
            tx_id=tx_id,
            payment_provider=self.payment_provider,
        )

        if get_tx_data is True:
            try:
                tx.set_transaction_data(get_raw_data=get_raw_data)
            except (
                NumConfirmationsNotMet,
                NoTXFound,
                NewConnectionError,
                HTTPError,
            ) as ne:
                logging.warning(ne)
                exit(1)
            except NoTXToProcess as e:
                logging.info(e)
                exit(1)

        return tx



from datetime import datetime
from decimal import Decimal
from ..exceptions import ExchangeRateDataError
from unittest import TestCase
from unittest.mock import MagicMock, patch

from apprisetransactions import settings, utils
from apprisetransactions.transactions import MoneroTransaction
from apprisetransactions.utils import currency_converter, parse_placeholders


class TestCase_Utils(TestCase):
    amount_in_usd_result = {
        "Response": "Success",
        "Type": 100,
        "Aggregated": False,
        "Data": [
            {
                "time": 1517245200,
                "close": 317.22,
                "high": 318.22,
                "low": 315.85,
                "open": 317.88,
                "volumefrom": 480.8,
                "volumeto": 152627.25,
            },
            {
                "time": 1517248800,
                "close": 318.49,
                "high": 318.79,
                "low": 316.04,
                "open": 317.22,
                "volumefrom": 965.1,
                "volumeto": 306274.66,
            },
        ],
        "TimeTo": 1517248800,
        "TimeFrom": 1517245200,
        "FirstValueInArray": True,
        "ConversionType": {"type": "direct", "conversionSymbol": ""},
        "RateLimit": {},
        "HasWarning": False,
    }
    amount_in_error = {
        "Response": "Error",
        "Message": "There is no data for the toSymbol PWD .",
        "HasWarning": False,
        "Type": 2,
        "RateLimit": {},
        "Data": {},
        "ParamWithError": "tsym",
    }

    def setUp(self):
        settings.init()
        self.transaction = MoneroTransaction(
            tx_id="4ea70add5d0c7db33557551b15cd174972fcfc73bf0f6a6b47b7837564b708d3",
            payment_provider="Monero",
            amount=Decimal("4.000000000000"),
            fee=Decimal("0.000962550000"),
            note="",
            recipient="9tQoHWyZ4yXUgbz9nvMcFZUfDy5hxcdZabQCxmNCUukKYicXegsDL7nQpcUa3A1pF6K3fhq3scsyY88tdB1MqucULcKzWZC",
            timestamp=datetime(2018, 1, 29, 13, 17, 18),
            confirmations=1,
        )

    def test_parse_placeholders(self):
        # simple
        body = "Check%20your%20wallet%20for%20more%20details"
        title = "New%20%7Bcurrency%7D%20received"
        title_expected = "New XMR received"
        body_expected = "Check your wallet for more details"
        body_after, title_after = parse_placeholders(self.transaction, body, title)
        self.assertEqual(body_after, body_expected)
        self.assertEqual(title_after, title_expected)

        # all transaction attributes
        body = "tx_id%3A%7Btx_id%7D%2Cpayment_provider" \
               "%3A%7Bpayment_provider%7D%2Camount%3A%7B" \
               "amount%7D%2Cfee%3A%7Bfee%7D"
        title = "note%3A%7Bnote%7D%2Crecipient%3A%7Brecipient" \
                "%7D%2Ctimestamp%3A%7Btimestamp%7D%2Cconfirmations%3A%7Bconfirmations%7D"
        title_expected = "note:,recipient:9tQoHWyZ4yXUgbz9nvMcFZUfDy5hxcdZabQCxmNC" \
                         "UukKYicXegsDL7nQpcUa3A1pF6K3fhq3scsyY88tdB1MqucULcKzWZC," \
                         "timestamp:2018-01-29 13:17:18,confirmations:1"
        body_expected = "tx_id:4ea70add5d0c7db33557551b15cd174972fcfc73bf0f6a6b47b7837564b708d3," \
                        "payment_provider:Monero,amount:4,fee:0.00096255"
        body_after, title_after = parse_placeholders(self.transaction, body, title)
        self.assertEqual(body_after, body_expected)
        self.assertEqual(title_after, title_expected)

        # case when the user specifies a placeholder, but that transaction's attribute was never populated with data
        def side_effect_lonely_tx(log):
            if mock_logger.debug.call_count == 1:
                self.assertEqual(
                    log,
                    'PaymentProvider: Monero TXID: lonelytx Status: The placeholder "{amount}" was specified, '
                    "but no data was found. Network error or user error. Action: Continuing",
                )
            if mock_logger.debug.call_count == 2:
                self.assertEqual(
                    log,
                    'PaymentProvider: Monero TXID: lonelytx Status: The placeholder "{recipient}" was specified, '
                    "but no data was found. Network error or user error. Action: Continuing",
                )

        transaction_with_no_details = MoneroTransaction(
            "lonelytx", payment_provider="Monero"
        )
        with patch.object(utils, "logging") as mock_logger:
            mock_logger.debug = MagicMock(side_effect=side_effect_lonely_tx)
            parse_placeholders(transaction_with_no_details, "{amount}", "{recipient}")

    @patch("apprisetransactions.utils.requests.get")
    def test_currency_converter(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = self.amount_in_usd_result
        amount_in_usd = currency_converter(
            self.transaction.amount,
            self.transaction.currency,
            "USD",
            self.transaction.timestamp,
        )
        self.assertEqual(amount_in_usd, Decimal("1269.660000000000081854523160"))

    @patch("apprisetransactions.utils.requests.get")
    def test_currency_converter_error(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = self.amount_in_error
        with self.assertRaises(ExchangeRateDataError):
            currency_converter(
                self.transaction.amount,
                self.transaction.currency,
                "ABCD",
                self.transaction.timestamp,
            )

    @patch("apprisetransactions.utils.requests.get")
    def test_parse_currency_conversion(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = self.amount_in_usd_result
        # simple
        body = "Congrats%21%20You%27ve%20received%20%24%7Bamount_in_usd%7D"
        title = "%24%7Bamount_in_usd%7D%20received%20via%20%7Bpayment_provider%7D"
        title_expected = "$1269.66 received via Monero"
        body_expected = "Congrats! You've received $1269.66"
        body_after, title_after = parse_placeholders(self.transaction, body, title)
        self.assertEqual(body_after, body_expected)
        self.assertEqual(title_after, title_expected)



import json
import os


class ServerConfig:
    protocol = ""
    host = ""
    port = ""
    path = ""
    user = ""
    password = ""
    access_token = ""
    default = False

    def __init__(self, config_file: str, **kwargs):
        self.config_file = config_file
        self.protocol = kwargs.pop("protocol", self.protocol)
        self.host = kwargs.pop("host", self.host)
        self.port = kwargs.pop("port", self.port)
        self.path = kwargs.pop("path", self.path)
        self.user = kwargs.pop("user", self.user)
        self.password = kwargs.pop("password", self.password)
        self.access_token = kwargs.pop("access_token", self.access_token)
        self.set_config()

    def set_config(self):
        if self.config_file is not None:
            if os.path.isfile(self.config_file):
                with open(self.config_file) as infile:
                    self.__dict__ = json.load(infile)
            else:
                self.default = True
        else:
            self.default = True



import json
from datetime import datetime
from decimal import Decimal
from ..exceptions import NoRawTXData
from unittest import TestCase
from unittest.mock import MagicMock, patch
from xml.dom.minidom import parseString

from dicttoxml import dicttoxml

from apprisetransactions import settings
from apprisetransactions.transactions import MoneroTransaction


class TestCase_Notify(TestCase):
    raw_data = {
        "transfer": {
            "address": "599BXkvzAPeD6EojXjW99gGeE19rxuS4g99o5xevp3jTFQYTNtSyYtqWAt2jg9MTP8aLTJKiuCJXg1Ro6xLtEUEm9rWHL7V",
            "amount": 100000000,
            "amounts": [100000000],
            "confirmations": 21,
            "double_spend_seen": False,
            "fee": 1006740000,
            "height": 597061,
            "locked": False,
            "note": "",
            "payment_id": "0000000000000000",
            "subaddr_index": {"major": 0, "minor": 0},
            "subaddr_indices": [{"major": 0, "minor": 0}],
            "suggested_confirmations_threshold": 1,
            "timestamp": 1591546700,
            "txid": "d0fb667f2975f79495e973d1437200f7b6c464956d33cd89289411e07a8c0b3a",
            "type": "in",
            "unlock_time": 0,
        },
        "transfers": [
            {
                "address": "599BXkvzAPeD6EojXjW99gGeE19rxuS4g99o5xevp3jTFQYTNtSyYtqWAt2jg9MTP8aLTJKiuCJXg1Ro6xLtEUEm9rWHL7V",
                "amount": 100000000,
                "amounts": [100000000],
                "confirmations": 21,
                "double_spend_seen": False,
                "fee": 1006740000,
                "height": 597061,
                "locked": False,
                "note": "",
                "payment_id": "0000000000000000",
                "subaddr_index": {"major": 0, "minor": 0},
                "subaddr_indices": [{"major": 0, "minor": 0}],
                "suggested_confirmations_threshold": 1,
                "timestamp": 1591546700,
                "txid": "d0fb667f2975f79495e973d1437200f7b6c464956d33cd89289411e07a8c0b3a",
                "type": "in",
                "unlock_time": 0,
            }
        ],
    }

    def setUp(self):
        settings.init()
        self.transaction = MoneroTransaction(
            tx_id="4ea70add5d0c7db33557551b15cd174972fcfc73bf0f6a6b47b7837564b708d3",
            payment_provider="Monero",
            amount=Decimal("4.000000000000"),
            fee=Decimal("0.000962550000"),
            note="",
            recipient="9tQoHWyZ4yXUgbz9nvMcFZUfDy5hxcdZabQCxmNCUukKYicXegsDL7nQpcUa3A1pF6K3fhq3scsyY88tdB1MqucULcKzWZC",
            timestamp=datetime(2018, 1, 29, 13, 17, 18),
            confirmations=1,
            _raw_data=self.raw_data,
        )

    @patch("apprisetransactions.transactions.transaction.Apprise")
    def test_notify_json(self, mock_apprise):
        def notify_json_sideeffect(json_body):
            self.assertEqual(json_body, json.dumps(self.raw_data))
            return True

        mock_apprise.notify = MagicMock(side_effect=notify_json_sideeffect)
        mock_apprise.return_value = mock_apprise
        self.assertTrue(self.transaction.notify(urls=["json://blah", "sns://blah"]))
        self.assertEqual(mock_apprise.notify.call_count, 1)

    def test_notify_exception(self):

        with self.assertRaises(NoRawTXData):
            MoneroTransaction(tx_id="txexample").notify(
                urls=["json://blah", "sns://blah"]
            )

        with self.assertRaises(NoRawTXData):
            MoneroTransaction(tx_id="txexample").notify(urls=["xml://blah"])

    @patch("apprisetransactions.transactions.transaction.Apprise")
    def test_notify_xml(self, mock_apprise):
        def notify_xml_sideeffect(xml_body):
            self.assertEqual(
                xml_body, parseString(dicttoxml(self.raw_data)).toprettyxml()
            )
            return True

        mock_apprise.notify = MagicMock(side_effect=notify_xml_sideeffect)
        mock_apprise.return_value = mock_apprise
        self.assertTrue(self.transaction.notify(urls=["xml://blah"]))
        self.assertEqual(mock_apprise.notify.call_count, 1)

    @patch("apprisetransactions.transactions.transaction.Apprise")
    def test_notify_other(self, mock_apprise):
        pass

        def notify_sideeffect(body, title, attach):
            self.assertEqual(body, "New XMR received")
            self.assertEqual(title, "Private Notification")
            return True

        mock_apprise.notify = MagicMock(side_effect=notify_sideeffect)
        mock_apprise.return_value = mock_apprise
        self.assertTrue(
            self.transaction.notify(
                urls=["pbul://asdfasdf"],
                body="New {currency} received",
                title="Private Notification",
            )
        )
        self.assertEqual(mock_apprise.notify.call_count, 1)

    @patch("apprisetransactions.transactions.transaction.parse_placeholders")
    @patch("apprisetransactions.transactions.transaction.Apprise")
    def test_znotify_other(self, mock_apprise, mock_parse_placeholders):
        def notify_sideeffect(body, title, attach):
            self.assertEqual(body, "New XMR received")
            self.assertEqual(title, "Private Notification")
            return True

        # prase_placeholders is tested in test_utils
        mock_parse_placeholders.return_value = (
            "New XMR received",
            "Private Notification",
        )
        mock_apprise.notify = MagicMock(side_effect=notify_sideeffect)
        mock_apprise.return_value = mock_apprise
        self.assertTrue(
            self.transaction.notify(
                urls=["pbul://asdfasdf"],
                body="New {currency} received",
                title="Private Notification",
            )
        )
        self.assertEqual(mock_apprise.notify.call_count, 1)
        self.assertEqual(mock_parse_placeholders.call_count, 1)



import logging

from . import settings


def backoff_confirmations_hdlr(details: []):
    tx = details["args"][0]
    logging.info(
        f"PaymentProvider: {tx.payment_provider} TXID: {tx.tx_id} "
        f"Status: Pending Network Confirmations Security_Level: {settings.security_level} Action: Retry #{details['tries']}"
    )


def backoff_connection_hdlr(details: []):
    tx = details["args"][0]

    logging.error(
        f"PaymentProvider: {tx.payment_provider} TXID: {tx.tx_id} "
        f"Status: No Connection to Monero Wallet RPC Action: Retry #{details['tries']}"
    )


def backoff_http_hdlr(details: []):
    tx = details["args"][0]
    logging.error(
        f"PaymentProvider: {tx.payment_provider} TXID: {tx.tx_id} "
        f"Status: Monero Wallet RPC request failed Action: Retry #{details['tries']}"
    )



__title__ = "apprisetransactions"
__version__ = "1.0.2"
__license__ = "MIT"
__status__ = "Production"

from .transactions import MoneroTransaction
from .factories import MoneroFactory
from .configuration import ServerConfig

# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

__all__ = [
    # Core
    "MoneroFactory",
    "MoneroTransaction",
    "ServerConfig",
]



# flake8: noqa: F401
from .tx_exceptions import (
    ExchangeRateDataError,
    NoTXFound,
    NumConfirmationsNotMet,
    NoTXToProcess,
    NoRawTXData,
)
from .apprise_exceptions import AppriseNotificationFailure



from ..transactions.transaction import Transaction
from monero.wallet import Wallet
from monero.backends.jsonrpc import JSONRPCWallet
from .. import settings
import backoff
from requests.exceptions import HTTPError
from urllib3.exceptions import ConnectionError
from ..backoff_hdlrs import (
    backoff_confirmations_hdlr,
    backoff_connection_hdlr,
    backoff_http_hdlr,
)
from monero.transaction import IncomingPayment
from ..exceptions.tx_exceptions import NoTXFound, NumConfirmationsNotMet, NoTXToProcess


def look_up_max_time() -> int:
    return settings.security_level * 600


# This object shouldn't be used directly, use it's associated factory to generate
class MoneroTransaction(Transaction):
    currency = "XMR"

    def __init__(self, tx_id: str, **kwargs):
        super().__init__(tx_id=tx_id, currency=self.currency, **kwargs)

    @backoff.on_exception(
        backoff.constant, HTTPError, max_time=60, on_backoff=backoff_http_hdlr
    )
    @backoff.on_exception(
        backoff.expo, ConnectionError, max_time=60, on_backoff=backoff_connection_hdlr,
    )
    @backoff.on_exception(
        backoff.constant,
        NumConfirmationsNotMet,
        max_time=look_up_max_time,
        interval=250,
        on_backoff=backoff_confirmations_hdlr,
    )
    # flake8: noqa: C901
    def set_transaction_data(self, get_raw_data: bool = False):
        if self.server_config.default is True:
            jsonrpcwallet = JSONRPCWallet()
        else:
            jsonrpcwallet = JSONRPCWallet(
                protocol=self.server_config.protocol,
                host=self.server_config.host,
                port=self.server_config.port,
                path=self.server_config.path,
                user=self.server_config.user,
                password=self.server_config.password,
            )

        try:
            wallet = Wallet(jsonrpcwallet)
        except Exception as e:
            # for some reason connection error isn't being caught
            if e.__class__.__name__ == "ConnectionError":
                raise ConnectionError
            else:
                raise e

        this_payment = None
        if settings.security_level == 0 or settings.security_level == -1:
            # get transaction in mem_pool
            incoming_payment = wallet.incoming(tx_id=self.tx_id, unconfirmed=True, confirmed=False)
            # https://github.com/monero-ecosystem/monero-python/issues/65
            if incoming_payment:
                tx_in_mem_pool = True
                this_payment: IncomingPayment = incoming_payment.pop()
            else:
                tx_in_mem_pool = False

            if tx_in_mem_pool is False and settings.security_level == 0:
                # raising, because we weren't able to find anything in the mem_pool
                # and security_level=0, which means that we only care about tx's in the mem_pool
                raise NoTXToProcess(
                    f"PaymentProvider: {self.payment_provider} TXID: {self.tx_id} "
                    "Status: No transaction found in mem_pool. "
                    "Your tx probably was processed and has been added to a block. "
                    "This is fine. "
                    f" Security_Level: {settings.security_level} "
                    "Action: Nothing"
                )

        if settings.security_level >= 1 or (
            settings.security_level == -1 and this_payment is None
        ):
            incoming_payments = wallet.incoming(tx_id=self.tx_id)
            if incoming_payments == []:
                raise NoTXFound(
                    f"Currency: {self.currency} TXID: {self.tx_id} "
                    "Status: No transaction found. Your tx probably hasn't been added to a block yet. "
                    "This is fine. "
                    f"Another notification will execute. Action: Nothing"
                )
            this_payment: IncomingPayment = incoming_payments[0]
            if settings.security_level > 1:

                if this_payment.transaction.confirmations < settings.security_level:
                    raise NumConfirmationsNotMet(
                        f"Currency: {self.currency} TXID: {self.tx_id} "
                        f"Confirmations: current {this_payment.transaction.confirmations}, "
                        f"expected {settings.security_level}"
                    )

        self.amount = this_payment.amount
        self.fee = this_payment.transaction.fee
        self.note = this_payment.note
        self.recipient = this_payment.local_address
        self.timestamp = this_payment.transaction.timestamp
        # if the tx is in the mem_pool confirmations will be null, set it to 0 instead
        self.confirmations = (
            this_payment.transaction.confirmations
            if this_payment.transaction.confirmations is not None
            else 0
        )

        if get_raw_data is True:
            self._raw_data = jsonrpcwallet.raw_request(
                "get_transfer_by_txid", {"txid": self.tx_id}
            )



from abc import ABC, abstractmethod
from ..configuration import ServerConfig
from decimal import Decimal
from datetime import datetime
from ..exceptions import NoRawTXData, AppriseNotificationFailure
from ..utils import parse_placeholders
import re
from xml.dom.minidom import parseString
from dicttoxml import dicttoxml
from apprise import Apprise
import json


# This object shouldn't be used directly, use it's associated factory to generate
class Transaction(ABC):
    tx_id = None
    currency: str = None
    payment_provider: str = None
    amount: Decimal = None
    fee: Decimal = None
    note: str = None
    recipient: str = None
    timestamp: datetime = None
    confirmations: int = None
    server_config: ServerConfig = None
    _raw_data: dict = None

    def __init__(self, **kwargs):
        self.tx_id = kwargs.pop("tx_id", self.tx_id)
        self.currency = kwargs.pop("currency", self.currency)
        self.payment_provider = kwargs.pop("payment_provider", self.payment_provider)
        self.amount = kwargs.pop("amount", self.amount)
        self.fee = kwargs.pop("fee", self.fee)
        self.note = kwargs.pop("note", self.note)
        self.recipient = kwargs.pop("recipient", self.recipient)
        self.timestamp = kwargs.pop("timestamp", self.timestamp)
        self.confirmations = kwargs.pop("confirmations", self.confirmations)
        self.server_config = kwargs.pop("server_config", self.server_config)
        self._raw_data = kwargs.pop("_raw_data", self._raw_data)

    # raw_data should be true when machine to machine communication is intended
    # raw_data will contain the response received from the server in it's entirety
    @abstractmethod
    def set_transaction_data(
        self, server_config: ServerConfig, get_raw_data: bool = False
    ):
        pass

    # flake8: noqa: F401
    def notify(
        self, urls: list, body: str = None, title: str = None, attach: str = None
    ) -> bool:
        # body/title are ignored for xml/json/sns urls
        if body is not None or title is not None:
            new_body, new_title = parse_placeholders(self, body, title)

        m2m_xml_urls = []
        m2m_json_urls = []
        urls_other = []
        while urls:
            url = urls.pop()
            if url[:3] == "xml":
                m2m_xml_urls.append(url)
                continue
            if url[:4] == "json":
                m2m_json_urls.append(url)
                continue
            if url[:3] == "sns":
                # sns can be used to send sms
                # but can also be used for m2m json message passing
                # txt messages shouldn't have json parsing so we want to separate them out
                phone_pattern = r"\+\d\d{3}\d{3}\d{4}"
                if re.search(phone_pattern, url) is None:
                    m2m_json_urls.append(url)
                    continue
            urls_other.append(url)

        # If apprise fails it will return False
        apprise_xml_result = True
        apprise_json_result = True
        apprise_other_result = True

        if m2m_xml_urls != []:
            if self._raw_data is None:
                raise NoRawTXData(
                    f"Currency: {self.currency} TXID: {self.tx_id} "
                    "Status: Raw TX Data Not Found Action: Report Bug"
                )
            xml: bytes = dicttoxml(self._raw_data)
            xml_body: str = parseString(xml).toprettyxml()
            apprise_xml_result = Apprise(servers=m2m_xml_urls).notify(xml_body)

        if m2m_json_urls != []:
            if self._raw_data is None:
                raise NoRawTXData(
                    f"Currency: {self.currency} TXID: {self.tx_id} "
                    "Status: Raw TX Data Not Found Action: Report Bug"
                )
            json_body = json.dumps(self._raw_data)
            apprise_json_result = Apprise(servers=m2m_json_urls).notify(json_body)

        if urls_other != []:
            apprise_other_result = Apprise(servers=urls_other).notify(
                new_body, new_title, attach=attach
            )

        if not apprise_json_result:
            raise AppriseNotificationFailure(
                f"PaymentProvider: {self.payment_provider} TXID: {self.tx_id} "
                f"Status: Apprise Failed To Send M2M JSON/SNS Notification Action: User Action required"
            )
        if not apprise_xml_result:
            raise AppriseNotificationFailure(
                f"PaymentProvider: {self.payment_provider} TXID: {self.tx_id} "
                f"Status: Apprise Failed To Send M2M XML Notification Action: User Action required"
            )
        if not apprise_other_result:
            raise AppriseNotificationFailure(
                f"PaymentProvider: {self.payment_provider} TXID: {self.tx_id} "
                f"Status: Apprise Failed To Send Notification Action: User Action required"
            )

        return apprise_json_result & apprise_xml_result & apprise_other_result



import json
from unittest import TestCase

from apprisetransactions import settings
from apprisetransactions.configuration import ServerConfig


class TestCase_Config(TestCase):
    def setUp(self):
        settings.init()
        self.config = ServerConfig(
            config_file="server.cfg",
            protocol="http",
            host="127.0.0.1",
            port=18088,
            path="/json_rpc",
            user="test",
            password="test",
        )

    def test_config_awrite(self):
        with open(self.config.config_file, "w") as outfile:
            json.dump(self.config.__dict__, outfile)
        pass

    def test_config_read(self):
        read_config = ServerConfig(config_file="server.cfg")
        self.assertEqual(read_config.protocol, self.config.protocol)
        self.assertEqual(read_config.__dict__, self.config.__dict__)



import logging
from abc import ABC, abstractmethod

from ..configuration import ServerConfig
from ..transactions.transaction import Transaction


class TransactionFactory(ABC):
    payment_provider: str = None
    server_config: ServerConfig = None

    def __init__(self, server_config_file: str = None):
        try:
            self.server_config = ServerConfig(config_file=server_config_file)
        except Exception:
            logging.critical(
                f"PaymentProvider: {self.payment_provider} Status: Failed to set server "
                f"config Action: Further debugging needed",
            )

    @abstractmethod
    def get_transaction(
        self, tx_id: str, get_tx_data: bool, get_raw_data: bool = False
    ) -> Transaction:
        pass



import argparse
import logging
import logging.config
from logging import Logger

from .utils import split_urls

from . import settings
from .factories import MoneroFactory
from pathlib import Path, PurePath


settings.init()
path = str(PurePath(Path(__file__).parent.absolute(), settings.logging_config_file))
logging.config.fileConfig(path)
logger: Logger = logging.getLogger(__name__)


def add_args():
    try:
        parser = argparse.ArgumentParser(description="Publish transactions")
        requiredNamed = parser.add_argument_group("required named arguments")
        requiredNamed.add_argument(
            "-p",
            "--payment_provider",
            dest="payment_provider",
            type=str,
            help="Provide the payment provider that you're notifying on",
            required=True,
        )
        requiredNamed.add_argument(
            "-x",
            "--tx_id",
            dest="tx_id",
            type=str,
            help="Pass %s, for tx-notify",
            required=True,
        )
        requiredNamed.add_argument(
            "-u",
            "--urls",
            dest="urls",
            type=str,
            help="Specify the apprise url(s) that you would like to use for notification(s), "
            "use a comma for multiple"
            "Pushbullet example: pbul://o.gn5kj6nfhv736I7jC3cj3QLRiyhgl98b"
            "E-mail example: mailto://myuserid:mypass@gmail.com",
            required=True,
        )
        parser.add_argument(
            "-b",
            "--body",
            dest="body",
            type=str,
            help="Specify the url encoded body/msg of the notification you would like to receive, "
            "hint: try {amount_in_usd} placeholder to convert to fiat"
            "hint #2: Go to https://www.urlencoder.org/ for easy url encoding",
            default="Check%20your%20wallet%20for%20more%20details",
        )
        parser.add_argument(
            "-t",
            "--title",
            dest="title",
            type=str,
            help="Specify the title of the notification you would like to receive"
            "This argument is ignored when using Amazon SNS.",
            default="Received%20transaction%20from%20%7Bpayment_provider%7D",
        )
        parser.add_argument(
            "-s",
            "--security_level",
            dest="security_level",
            type=int,
            help="-1 - notifies when tx is in mem_pool or block\n"
            "0 - notifies when tx is in mem_pool\n"
            "1:n - notifies when tx has been added to a block and has n confirmations",
            default=settings.BlockchainSecurity.MEM_POOL_ONLY,
        )
        parser.add_argument(
            "-c",
            "--server_config",
            dest="server_config_file",
            type=str,
            help="To retrieve additional data about the transaction, we need to query a server. See the README",
            default=None,
        )
        parser.add_argument(
            "-a",
            "--attach",
            dest="attach",
            type=str,
            help="To retrieve additional data about the transaction, we need to query a server. See the README",
            default=None,
        )
        parser.add_argument(
            "--get_tx_details",
            default=False,
            dest="get_tx_data",
            action="store_true",
            help="Specify this argument if you would like the details of the transaction "
            "I.e. amount, receiver, etc"
            "If you use placeholders such as {amount_in_usd} then you need to have this argument",
        )
        parser.add_argument(
            "--get_raw_tx_data",
            default=False,
            dest="get_raw_data",
            action="store_true",
            help="Specify this argument if you are forwarding the data to another server"
            " via json, xml, sns. "
            "You should be using a redis queue if you want the python object to be passed",
        )
        parser.add_argument(
            "--debug",
            default=False,
            dest="debug",
            action="store_true",
            help="Enable additional logging",
        )

        return parser.parse_args()

    except Exception as e:
        logger.exception(e)
        exit(1)


def main():
    args = add_args()
    logger.info(
        f"PaymentProvider: {args.payment_provider} TXID: {args.tx_id} "
        f"Status: Script executing Action: Processing"
    )

    if args.debug:
        logging.root.setLevel(logging.DEBUG)
    else:
        logging.root.setLevel(logging.INFO)

    settings.security_level = args.security_level

    payment_providers = {"Monero": MoneroFactory}

    if args.payment_provider in payment_providers:
        transaction_factory = payment_providers[args.payment_provider](
            server_config_file=args.server_config_file
        )
    else:
        logger.critical("The supplied payment provider is not supported")
        exit(1)

    try:
        transaction = transaction_factory.get_transaction(
            tx_id=args.tx_id,
            get_tx_data=args.get_tx_data,
            get_raw_data=args.get_raw_data,
        )
    except Exception as e:
        logger.exception(e)
        exit(1)

    try:
        apprise_result = transaction.notify(
            urls=split_urls(args.urls),
            body=args.body,
            title=args.title,
            attach=args.attach,
        )
        if apprise_result:
            logger.info(
                f"PaymentProvider: {transaction.payment_provider} TXID: {transaction.tx_id} "
                f"Status: Notification(s) sent via apprise"
            )
    except Exception as ae:
        logger.exception(ae)
        exit(1)



# settings.py
class BlockchainSecurity:
    MEM_POOL_ONLY = 0
    MEM_POOL_AND_BLOCKCHAIN = -1
    IN_A_BLOCK = 1
    FULLY_CONFIRMED = 10


def init():
    global security_level
    security_level = BlockchainSecurity.MEM_POOL_ONLY
    global server_config
    server_config = None
    global logging_config_file
    logging_config_file = "logging_config.ini"



import pathlib
from setuptools import setup
from setuptools import find_packages


HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="apprise-transactions",
    version="1.0.2",
    packages=find_packages(),
    include_package_data=True,
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/apprises/apprise-transactions",
    keywords="Push Notifications Alerts Apprise Email AWS SNS Boxcar ClickSend "
    "Discord Dbus Emby Faast Flock Gitter Gnome Gotify Growl IFTTT Join "
    "Kavenegar KODI Kumulos MacOS Mailgun Matrix Mattermost MessageBird "
    "MSG91 Nexmo Nextcloud Notica Notifico Office365 Prowl PushBullet "
    "Pushjet Pushed Pushover PushSafer Rocket.Chat Ryver SendGrid "
    "SimplePush Sinch Slack Stride Syslog Techulus Push Telegram Twilio "
    "Twist Twitter XBMC Monero MSTeams Windows Webex CLI API",
    license="MIT",
    author="Apprises",
    author_email="",
    description="Push Notifications that work with just about every payment platform!",
    install_requires=open("requirements.txt").readlines(),
    entry_points={
        "console_scripts": ["apprisetransactions=apprisetransactions.cli:main"]
    },
    tests_require=open("dev-requirements.txt").readlines(),
)



from datetime import datetime
from decimal import Decimal
from ..exceptions import NoTXFound, NoTXToProcess, NumConfirmationsNotMet
from unittest import TestCase
from unittest.mock import MagicMock, patch

from monero.account import Account
from monero.address import address
from monero.numbers import PaymentID
from monero.transaction import IncomingPayment, Transaction

from apprisetransactions import settings
from apprisetransactions.configuration import ServerConfig
from apprisetransactions.transactions import MoneroTransaction


class TestCase_MoneroTX(TestCase):
    raw_data = {
        "transfer": {
            "address": "599BXkvzAPeD6EojXjW99gGeE19rxuS4g99o5xevp3jTFQYTNtSyYtqWAt2jg9MTP8aLTJKiuCJXg1Ro6xLtEUEm9rWHL7V",
            "amount": 100000000,
            "amounts": [100000000],
            "confirmations": 21,
            "double_spend_seen": False,
            "fee": 1006740000,
            "height": 597061,
            "locked": False,
            "note": "",
            "payment_id": "0000000000000000",
            "subaddr_index": {"major": 0, "minor": 0},
            "subaddr_indices": [{"major": 0, "minor": 0}],
            "suggested_confirmations_threshold": 1,
            "timestamp": 1591546700,
            "txid": "d0fb667f2975f79495e973d1437200f7b6c464956d33cd89289411e07a8c0b3a",
            "type": "in",
            "unlock_time": 0,
        },
        "transfers": [
            {
                "address": "599BXkvzAPeD6EojXjW99gGeE19rxuS4g99o5xevp3jTFQYTNtSyYtqWAt2jg9MTP8aLTJKiuCJXg1Ro6xLtEUEm9rWHL7V",
                "amount": 100000000,
                "amounts": [100000000],
                "confirmations": 21,
                "double_spend_seen": False,
                "fee": 1006740000,
                "height": 597061,
                "locked": False,
                "note": "",
                "payment_id": "0000000000000000",
                "subaddr_index": {"major": 0, "minor": 0},
                "subaddr_indices": [{"major": 0, "minor": 0}],
                "suggested_confirmations_threshold": 1,
                "timestamp": 1591546700,
                "txid": "d0fb667f2975f79495e973d1437200f7b6c464956d33cd89289411e07a8c0b3a",
                "type": "in",
                "unlock_time": 0,
            }
        ],
    }

    class MockBackend(object):
        def __init__(self):
            self.transfers = []
            tx = Transaction(
                timestamp=datetime(2018, 1, 29, 15, 0, 25),
                height=1087606,
                hash="a0b876ebcf7c1d499712d84cedec836f9d50b608bb22d6cb49fd2feae3ffed14",
                fee=Decimal("0.00352891"),
            )
            pm = IncomingPayment(
                amount=Decimal("1"),
                local_address=address(
                    "Bf6ngv7q2TBWup13nEm9AjZ36gLE6i4QCaZ7XScZUKDUeGbYEHmPRdegKGwLT8tBBK7P6L32RELNzCR6QzNFkmogDjvypyV"
                ),
                payment_id=PaymentID(
                    "0166d8da6c0045c51273dd65d6f63734beb8a84e0545a185b2cfd053fced9f5d"
                ),
                transaction=tx,
            )
            self.transfers.append(pm)
            tx = Transaction(
                timestamp=datetime(2018, 1, 29, 14, 57, 47),
                height=1087601,
                hash="f34b495cec77822a70f829ec8a5a7f1e727128d62e6b1438e9cb7799654d610e",
                fee=Decimal("0.008661870000"),
            )
            pm = IncomingPayment(
                amount=Decimal("3.000000000000"),
                local_address=address(
                    "BhE3cQvB7VF2uuXcpXp28Wbadez6GgjypdRS1F1Mzqn8Advd6q8VfaX8ZoEDobjejrMfpHeNXoX8MjY8q8prW1PEALgr1En"
                ),
                payment_id=PaymentID("f75ad90e25d71a12"),
                transaction=tx,
            )
            self.transfers.append(pm)
            tx = Transaction(
                timestamp=datetime(2018, 1, 29, 13, 17, 18),
                height=1087530,
                hash="5c3ab739346e9d98d38dc7b8d36a4b7b1e4b6a16276946485a69797dbf887cd8",
                fee=Decimal("0.000962550000"),
            )
            pm = IncomingPayment(
                amount=Decimal("10.000000000000"),
                local_address=address(
                    "9tQoHWyZ4yXUgbz9nvMcFZUfDy5hxcdZabQCxmNCUukKYicXegsDL7nQpcUa3A1pF6K3fhq3scsyY88tdB1MqucULcKzWZC"
                ),
                payment_id=PaymentID("f75ad90e25d71a12"),
                transaction=tx,
            )
            self.transfers.append(pm)
            tx = Transaction(
                timestamp=datetime(2018, 1, 29, 13, 17, 18),
                height=1087608,
                hash="4ea70add5d0c7db33557551b15cd174972fcfc73bf0f6a6b47b7837564b708d3",
                fee=Decimal("0.000962550000"),
                confirmations=1,
            )
            pm = IncomingPayment(
                amount=Decimal("4.000000000000"),
                local_address=address(
                    "9tQoHWyZ4yXUgbz9nvMcFZUfDy5hxcdZabQCxmNCUukKYicXegsDL7nQpcUa3A1pF6K3fhq3scsyY88tdB1MqucULcKzWZC"
                ),
                payment_id=PaymentID("f75ad90e25d71a12"),
                transaction=tx,
            )
            self.transfers.append(pm)
            tx = Transaction(
                timestamp=datetime(2018, 1, 29, 13, 17, 18),
                height=1087530,
                hash="e9a71c01875bec20812f71d155bfabf42024fde3ec82475562b817dcc8cbf8dc",
                fee=Decimal("0.000962550000"),
            )
            pm = IncomingPayment(
                amount=Decimal("2.120000000000"),
                local_address=address(
                    "9tQoHWyZ4yXUgbz9nvMcFZUfDy5hxcdZabQCxmNCUukKYicXegsDL7nQpcUa3A1pF6K3fhq3scsyY88tdB1MqucULcKzWZC"
                ),
                payment_id=PaymentID("cb248105ea6a9189"),
                transaction=tx,
            )
            self.transfers.append(pm)
            tx = Transaction(
                timestamp=datetime(2018, 1, 29, 14, 57, 47),
                height=1087601,
                hash="5ef7ead6a041101ed326568fbb59c128403cba46076c3f353cd110d969dac808",
                fee=Decimal("0.000962430000"),
            )
            pm = IncomingPayment(
                amount=Decimal("7.000000000000"),
                local_address=address(
                    "BhE3cQvB7VF2uuXcpXp28Wbadez6GgjypdRS1F1Mzqn8Advd6q8VfaX8ZoEDobjejrMfpHeNXoX8MjY8q8prW1PEALgr1En"
                ),
                payment_id=PaymentID("0000000000000000"),
                transaction=tx,
            )
            self.transfers.append(pm)
            tx = Transaction(
                timestamp=datetime(2018, 1, 29, 13, 17, 18),
                height=1087606,
                hash="cc44568337a186c2e1ccc080b43b4ae9db26a07b7afd7edeed60ce2fc4a6477f",
                fee=Decimal("0.000962550000"),
            )
            pm = IncomingPayment(
                amount=Decimal("10.000000000000"),
                local_address=address(
                    "9tQoHWyZ4yXUgbz9nvMcFZUfDy5hxcdZabQCxmNCUukKYicXegsDL7nQpcUa3A1pF6K3fhq3scsyY88tdB1MqucULcKzWZC"
                ),
                payment_id=PaymentID("0000000000000000"),
                transaction=tx,
            )
            self.transfers.append(pm)
            tx = Transaction(
                timestamp=datetime(2018, 1, 29, 21, 13, 28),
                height=None,
                hash="d29264ad317e8fdb55ea04484c00420430c35be7b3fe6dd663f99aebf41a786c",
                fee=Decimal("0.000961950000"),
            )
            pm = IncomingPayment(
                amount=Decimal("3.140000000000"),
                local_address=address(
                    "9tQoHWyZ4yXUgbz9nvMcFZUfDy5hxcdZabQCxmNCUukKYicXegsDL7nQpcUa3A1pF6K3fhq3scsyY88tdB1MqucULcKzWZC"
                ),
                payment_id=PaymentID("03f6649304ea4cb2"),
                transaction=tx,
            )
            self.transfers.append(pm)

        def height(self):
            return 1087607

        def accounts(self):
            return [Account(self, 0)]

        def transfers_in(self, account, pmtfilter):
            return list(pmtfilter.filter(self.transfers))

    @patch("apprisetransactions.transactions.monero_transaction.JSONRPCWallet")
    def test_set_transaction_data(self, mock_backend):
        transaction: MoneroTransaction = MoneroTransaction(
            tx_id="a0b876ebcf7c1d499712d84cedec836f9d50b608bb22d6cb49fd2feae3ffed14",
            server_config=ServerConfig(config_file=None),
        )
        settings.security_level = 1
        mock_backend.side_effect = self.MockBackend
        transaction.set_transaction_data(get_raw_data=False)

        self.assertEqual(
            transaction.tx_id,
            "a0b876ebcf7c1d499712d84cedec836f9d50b608bb22d6cb49fd2feae3ffed14",
        )
        self.assertEqual(transaction.currency, "XMR")
        # no payment provider because we didn't go through the factory
        self.assertEqual(transaction.payment_provider, None)
        self.assertEqual(transaction.amount, Decimal("1"))
        self.assertEqual(transaction.fee, Decimal("0.00352891"))
        self.assertEqual(transaction.note, "")
        self.assertEqual(
            transaction.recipient,
            "Bf6ngv7q2TBWup13nEm9AjZ36gLE6i4QCaZ7XScZUKD"
            "UeGbYEHmPRdegKGwLT8tBBK7P6L32RELNzCR6QzNFkmogDjvypyV",
        )
        self.assertEqual(transaction.timestamp, datetime(2018, 1, 29, 15, 0, 25))
        self.assertEqual(transaction.confirmations, 0)

    @patch("apprisetransactions.transactions.monero_transaction.JSONRPCWallet")
    def test_set_transaction_data_mem_pool(self, mock_backend):
        transaction: MoneroTransaction = MoneroTransaction(
            tx_id="d29264ad317e8fdb55ea04484c00420430c35be7b3fe6dd663f99aebf41a786c",
            server_config=ServerConfig(config_file=None),
        )
        settings.security_level = 0
        mock_backend.side_effect = self.MockBackend
        transaction.set_transaction_data(get_raw_data=False)

        self.assertEqual(
            transaction.tx_id,
            "d29264ad317e8fdb55ea04484c00420430c35be7b3fe6dd663f99aebf41a786c",
        )
        self.assertEqual(transaction.currency, "XMR")
        # no payment provider because we didn't go through the factory
        self.assertEqual(transaction.payment_provider, None)
        self.assertEqual(transaction.amount, Decimal("3.140000000000"))
        self.assertEqual(transaction.fee, Decimal("0.000961950000"))
        self.assertEqual(transaction.note, "")
        self.assertEqual(
            transaction.recipient,
            "9tQoHWyZ4yXUgbz9nvMcFZUfDy5hxcdZabQCxmNCUukKYicXegsDL7nQpcUa3A1pF6K3fhq3scsyY88tdB1MqucULcKzWZC",
        )
        self.assertEqual(transaction.timestamp, datetime(2018, 1, 29, 21, 13, 28))
        self.assertEqual(transaction.confirmations, 0)

    @patch("apprisetransactions.transactions.monero_transaction.JSONRPCWallet")
    def test_set_transaction_data_exception_numconfirmations(self, mock_backend):
        transaction: MoneroTransaction = MoneroTransaction(
            tx_id="4ea70add5d0c7db33557551b15cd174972fcfc73bf0f6a6b47b7837564b708d3",
            server_config=ServerConfig(config_file=None),
        )
        settings.security_level = 2
        mock_backend.side_effect = self.MockBackend
        with self.assertRaises(NumConfirmationsNotMet):
            transaction.set_transaction_data(get_raw_data=False)

    @patch("apprisetransactions.transactions.monero_transaction.Wallet")
    @patch("apprisetransactions.transactions.monero_transaction.JSONRPCWallet")
    def test_get_tx_data_exceptions(self, mock_backend, mock_wallet):
        transaction: MoneroTransaction = MoneroTransaction(
            tx_id="not_there", server_config=ServerConfig(config_file=None)
        )
        settings.security_level = -1
        mock_backend.return_value = MagicMock()
        mock_wallet.return_value.incoming.return_value = []
        with self.assertRaises(NoTXFound):
            transaction.set_transaction_data(get_raw_data=False)

    @patch("apprisetransactions.transactions.monero_transaction.JSONRPCWallet")
    def test_get_tx_data_exception_notxtoprocess(self, mock_backend):
        settings.security_level = 0
        # tx in_block, not in mem pool
        transaction: MoneroTransaction = MoneroTransaction(
            tx_id="a0b876ebcf7c1d499712d84cedec836f9d50b608bb22d6cb49fd2feae3ffed14",
            server_config=ServerConfig(config_file=None),
        )
        mock_backend.side_effect = self.MockBackend
        with self.assertRaises(NoTXToProcess):
            transaction.set_transaction_data(get_raw_data=False)

    @patch("apprisetransactions.transactions.monero_transaction.JSONRPCWallet")
    def test_set_raw_transaction_data(self, mock_backend):
        transaction: MoneroTransaction = MoneroTransaction(
            tx_id="d0fb667f2975f79495e973d1437200f7b6c464956d33cd89289411e07a8c0b3a",
            server_config=ServerConfig(config_file=None),
        )
        mock_backend.return_value.raw_request.return_value = self.raw_data
        transaction.set_transaction_data(get_raw_data=True)
        self.assertEqual(transaction._raw_data, self.raw_data)



import logging
import re
from datetime import datetime
from decimal import Decimal
from .exceptions import ExchangeRateDataError
from urllib import parse

import requests
from requests.exceptions import HTTPError


# Regular expression used to destinguish between multiple URLs
URL_DETECTION_RE = re.compile(
    r'([a-z0-9]+?:\/\/.*?)[\s,]*(?=$|[a-z0-9]+?:\/\/)', re.I)


# flake8: noqa: C901
def parse_placeholders(tx: object, body: str, title: str = None):
    amount_in_fiat = None
    fiat = None
    msg = [body, title]

    # url decode the message just in case the parameters passed are encoded
    i = 0
    for n in msg:
        if msg[i] is not None:
            msg[i] = parse.unquote(msg[i])
            i += 1

    # loop through all the properties of this object
    for prop in dir(tx):
        # we don't care about private and protected attributes
        if prop[0] == "_":
            continue
        # replace any placeholder attributes with the values from this object
        prop_re = f"{{{prop}}}"
        if (
            re.search(f".*{prop_re}.*", msg[0]) is None
            and re.search(f".*{prop_re}.*", msg[1]) is None
            and re.search("{amount_in_(...)}", msg[0]) is None
            and re.search("{amount_in_(...)}", msg[1]) is None
        ):
            continue

        attr = getattr(tx, prop)
        if attr is None:
            logging.debug(
                f"PaymentProvider: {tx.payment_provider} TXID: {tx.tx_id} "
                f'Status: The placeholder "{prop_re}" was specified, but no data was found.'
                f" Network error or user error. Action: Continuing"
            )
            continue

        i = 0
        for n in msg:
            if msg[i] is not None:
                if re.search(f".*{prop_re}.*", msg[i]):
                    # just swap the value with the {placeholder}
                    # .rstrip('0').rstrip('.') is to remove the trailing zeros from decimals i.e. amount
                    msg[i] = re.sub(
                        prop_re, attr.__str__().rstrip("0").rstrip("."), msg[i]
                    )
                if (
                    prop == "amount"
                    and re.search("{amount_in_(...)}", msg[i]) is not None
                ):
                    # convert amount to fiat if necessary
                    fiat = (
                        re.search("{amount_in_(...)}", msg[i]).group(1)
                        if fiat is None
                        else fiat.lower()
                    )
                    prop_fiat = f"{{amount_in_{fiat}}}"
                    fiat = fiat.upper()
                    # no need to make multiple requests for both body and title
                    if amount_in_fiat is None:
                        amount_in_fiat_unformatted = currency_converter(
                            tx.amount, tx.currency, fiat, tx.timestamp
                        )
                        amount_in_fiat = "{0:.2f}".format(amount_in_fiat_unformatted)

                    msg[i] = re.sub(prop_fiat, amount_in_fiat, msg[i])

            i += 1

    return msg[0], msg[1]  # body, title


def currency_converter(
    amount: Decimal, from_currency: str, to_currency: str, timestamp: datetime = None
):
    tx_datetime = timestamp if timestamp is not None else datetime.now()
    # convert to unix time
    tx_datetime = tx_datetime.timestamp()

    uri = "https://min-api.cryptocompare.com"
    path = "/data/histohour"
    queryStr = (
        f"?fsym={from_currency}&tsym={to_currency}&limit=1&e=CCCAGG&toTs={tx_datetime}"
    )
    url = uri + path + queryStr
    try:
        r = requests.get(url)
        r.raise_for_status()
        if r.json()["Response"] == "Error":
            raise ExchangeRateDataError(r.json()["Message"])
        hdp = r.json()["Data"][1]
    except HTTPError as e:
        raise HTTPError
    except KeyError as e:
        raise KeyError

    avgPrice = Decimal((hdp["high"] + hdp["low"]) / 2)
    return avgPrice * amount


def split_urls(urls):
    """
    Takes a string containing URLs separated by comma's and/or spaces and
    returns a list.
    """

    try:
        results = URL_DETECTION_RE.findall(urls)

    except TypeError:
        results = []

    if len(results) > 0 and results[len(results) - 1][-1] != urls[-1]:
        # we always want to save the end of url URL if we can; This handles
        # cases where there is actually a comma (,) at the end of a single URL
        # that would have otherwise got lost when our regex passed over it.
        results[len(results) - 1] += \
            re.match(r'.*?([\s,]+)?$', urls).group(1).rstrip()

    return results



from unittest import TestCase

from apprisetransactions.factories import MoneroFactory
from apprisetransactions.transactions import MoneroTransaction


class TestCase_MoneroFactory(TestCase):
    def setUp(self) -> None:
        self.monero_factory = MoneroFactory(server_config_file=None)

    def test_instantiation(self):
        self.assertTrue(self.monero_factory.server_config.default)
        self.assertEqual(MoneroFactory, type(self.monero_factory))
        self.assertIsNot(
            Exception,
            type(
                self.monero_factory.get_transaction(
                    tx_id="", get_tx_data=False, get_raw_data=True
                )
            ),
        )

    def test_get_transaction(self):
        transaction = self.monero_factory.get_transaction(
            tx_id="test", get_tx_data=False, get_raw_data=False
        )
        self.assertEqual(MoneroTransaction, type(transaction))



class ExchangeRateDataError(Exception):
    pass


class NoTXFound(Exception):
    pass


class NumConfirmationsNotMet(Exception):
    pass


class NoTXToProcess(Exception):
    pass


# something went wrong and no raw data about the tx was retrieved
class NoRawTXData(Exception):
    pass
